package org.recursed.w.client.place;

import org.recursed.w.client.place.PlaceChangeEvent.PlaceChangeHandler;

import com.google.gwt.core.client.GWT;
import com.google.gwt.event.logical.shared.ValueChangeEvent;
import com.google.gwt.event.logical.shared.ValueChangeHandler;
import com.google.gwt.event.shared.HandlerManager;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.Command;
import com.google.gwt.user.client.History;
import com.google.inject.Inject;
import com.google.inject.Singleton;

@Singleton
public class PlaceService implements ValueChangeHandler<String> {

    public interface Factory {
        /**
         * Given the current arguments on the history, resolve a Place. Note
         * that the arguments will be non-null or greater than zero.
         * 
         * @param urlShortCode
         * @return
         */
        public Place getPlace(String urlShortCode, String[] tokens);

        /**
         * When the application first fires up, or when a zero-argument place
         * change is detected, the Factory will be asked to provide a Place this
         * service will change to.
         * 
         * @return
         */
        public Place getDefaultPlace();

    }

    private Factory factory;
    private HandlerManager handler;

    @Inject
    private PlaceService(Factory factory) {
        this.factory = factory;
        this.handler = new HandlerManager(null);
        History.addValueChangeHandler(this);
    }

    public void fireInitialPlace() {
        String initialToken = History.getToken();
        if (initialToken == null || initialToken.length() == 0) {
            Place start = factory.getDefaultPlace();
            updateHistoryWithPlace(start);
        } else {
            processFragment(initialToken);
        }
    }

    public HandlerRegistration addPlaceChangeHandler(PlaceChangeHandler handler) {
        return this.handler.addHandler(PlaceChangeEvent.TYPE, handler);
    }

    public void changePlace(Place p) {
        updateHistoryWithPlace(p);
    }

    private String getStringForPlace(Place p) {
        String placeShortCode = p.getURLShortCode();
        StringBuilder result = new StringBuilder(16);
        result.append(placeShortCode);
        String[] tokens = p.getTokenizedArguments();
        if (tokens != null && tokens.length > 0) {
            for (String token : tokens) {
                result.append(':');
                result.append(token);
            }
        }
        return result.toString();
    }

    private void updateHistoryWithPlace(Place p) {
        String newHistoryItem = getStringForPlace(p);
        History.newItem(newHistoryItem, false);
        PlaceChangeEvent e = new PlaceChangeEvent(p);
        handler.fireEvent(e);
        registerPlaceInGoogleAnalytics(newHistoryItem);
    }

    @Override
    public void onValueChange(ValueChangeEvent<String> event) {
        String fullString = event.getValue();
        processFragment(fullString);
        registerPlaceInGoogleAnalytics(fullString);
    }

    private void processFragment(String fullString) {
        int firstColon = fullString.indexOf(':');
        Place p = null;
        if (firstColon > 0) {
            // we have arguments..
            String urlShortCode = fullString.substring(0, firstColon);
            String remainingArgs = fullString.substring(firstColon);
            String[] arguments = remainingArgs.split(":");
            p = factory.getPlace(urlShortCode, arguments);
        } else {
            // no arguments
            p = factory.getPlace(fullString, new String[] {});
        }
        final Place toFire = p;
        Command c = new Command() {

            @Override
            public void execute() {
                handler.fireEvent(new PlaceChangeEvent(toFire));
            }

        };
        p.inflateAndFire(c);
    }

    private void registerPlaceInGoogleAnalytics(String fragment) {
        try {
            registerPlaceInGoogleAnalyticsNative('/'+fragment);
        } catch (Exception e) {
            GWT.log("Failed to talk to GA", e);
        }
    }
    private native void registerPlaceInGoogleAnalyticsNative(String fragment) /*-{
        $wnd._gaq.push(['_trackPageview', fragment]);
    }-*/;

}
