package org.recursed.w.client.relationshipgraph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;

import org.recursed.w.client.characteristics.HasChartControls;
import org.recursed.w.client.place.PlaceService;
import org.recursed.w.client.place.WordDetailPlace;
import org.recursed.w.shared.WordDTO;

import com.google.gwt.event.dom.client.ClickEvent;
import com.google.gwt.event.dom.client.ClickHandler;
import com.google.gwt.event.shared.HandlerRegistration;
import com.google.gwt.user.client.ui.AbsolutePanel;
import com.google.gwt.user.client.ui.Label;
import com.google.gwt.user.client.ui.RootPanel;
import com.google.gwt.widgetideas.graphics.client.Color;
import com.google.gwt.widgetideas.graphics.client.GWTCanvas;
import com.google.inject.Inject;

public class RelationshipGraphDisplay implements
        RelationshipGraphPresenter.Display {

    private static final String GRAPH_DIV = "graph";
    private static final String WORD_TITLE_STYLE = "word_title_style";
    private static final String LABEL_STYLE = "middle";
    private static final Color DK_GRAY = new Color("#BB2222");
    private static final Color[] SECONDARY = new Color[] {
        new Color("#553C31"),
        new Color("#EE5313"),
        new Color("#7f7f7f"),
        new Color("#54C354"),
        new Color("#60EE13"),
        new Color("#CA52F6")
    };

    private int width;
    private int height;
    private double middleX, middleY;
    
    private PlaceService placeService;
    private Label wordTitle;
    private GWTCanvas canvas;
    private AbsolutePanel absolutePanel;
    private HasChartControls controls;
    private ArrayList<double[]> phaseOneCoords;
    private ArrayList<double[]> phaseTwoCoords;
    private HashSet<WordDTO> secondaryWords;
    private ArrayList<WordDTO[]> secondaryTuples;
    private HashMap<WordDTO, Color> colors;
    private HashMap<Integer, double[]> coordinates;
    private int currentPhaseOneLinkage, currentPhaseTwoLinkage;
    private ArrayList<Label> labels = new ArrayList<Label>();
    private ArrayList<HandlerRegistration> registrations = new ArrayList<HandlerRegistration>();

    @Inject
    private RelationshipGraphDisplay(PlaceService ps) {
        width = 480;
        height = 480;
        this.placeService = ps;
        absolutePanel = new AbsolutePanel();
        canvas = new GWTCanvas(width, height);
        absolutePanel.add(canvas);
        RootPanel graphPanel = RootPanel.get(GRAPH_DIV);
        wordTitle = new Label();
        wordTitle.setStyleName(WORD_TITLE_STYLE);
        graphPanel.add(wordTitle);
        graphPanel.add(absolutePanel);
        createChartControls();
    }

    private void createChartControls() {
        controls = new HasChartControls() {

            @Override
            public void render() {
                // does nothing..
            }

            @Override
            public void clear() {
                canvas.clear();
                for (Label l : labels) {
                    absolutePanel.remove(l);
                }
                labels = new ArrayList<Label>();
                for (HandlerRegistration r : registrations) {
                    r.removeHandler();
                }
                registrations = new ArrayList<HandlerRegistration>();
            }
        };
    }

    @Override
    public HasChartControls getControls() {
        return controls;
    }

    @Override
    public void setCoreWord(WordDTO wordDTO) {
        middleX = width/2;
        middleY = height/2;
        coordinates = new HashMap<Integer, double[]>();
        phaseOneCoords = new ArrayList<double[]>();
        phaseTwoCoords = new ArrayList<double[]>();
        secondaryTuples = new ArrayList<WordDTO[]>();
        secondaryWords = new HashSet<WordDTO>();
        colors = new HashMap<WordDTO, Color>();
        computeCoords(phaseOneCoords, 100d, 6, 0);
        currentPhaseOneLinkage = 0;
        currentPhaseTwoLinkage = 0;
        drawWord(new double[] {middleX, middleY}, 10, wordDTO, DK_GRAY);
        wordTitle.setText(wordDTO.getDisplayString());
    }

    private void computeCoords(ArrayList<double[]> collection, double radius, int nodes, double offset) {
        double degreeSegments = (2d*Math.PI)/nodes;
        for (int x=0; x < nodes; x++) {
            double newX = radius * Math.cos(offset + (degreeSegments * x));
            double newY = radius * Math.sin(offset + (degreeSegments * x));
            collection.add(new double[]{middleX + newX, middleY + newY});
        }
    }

    @Override
    public void addSecondaryLinkage(WordDTO from, WordDTO to) {
        if (coordinates.get(to.getId()) == null) {
            // this is a new word, not in the inner ring
            secondaryWords.add(to);
        }
        secondaryTuples.add(new WordDTO[] {from, to});
    }

    @Override
    public void renderSecondaryLinkages() {
        computeCoords(phaseTwoCoords, 200d, secondaryWords.size(), 0.5);
        for (WordDTO toWord : secondaryWords) {
            double[] coords = phaseTwoCoords.get(currentPhaseTwoLinkage);
            drawWord(coords, 3, toWord, null);
            currentPhaseTwoLinkage++;
        }
        canvas.setLineWidth(1);
        for (WordDTO[] tuple : secondaryTuples) {
            Color c = colors.get(tuple[0]);
            canvas.setStrokeStyle(c);
            drawLinkage(tuple[0], tuple[1]);
        }
    }
    
    @Override
    public void addLinkage(WordDTO from, WordDTO to) {
        if (currentPhaseOneLinkage > phaseOneCoords.size()) {
            throw new IllegalArgumentException("Exceeded linkage count");
        }
        double[] coords = phaseOneCoords.get(currentPhaseOneLinkage);
        currentPhaseOneLinkage++;
        canvas.setStrokeStyle(DK_GRAY);
        canvas.setLineWidth(2);
        Color c = colors.get(to);
        if (c == null) {
            c = SECONDARY[colors.size()];
            colors.put(to, c);
        }
        drawWord(coords, 3, to, c);
        drawLinkage(from, to);
    }

    private void drawLinkage(WordDTO from, WordDTO to) {
        double[] fc = coordinates.get(from.getId());
        double[] tc = coordinates.get(to.getId());
        canvas.beginPath();
        canvas.moveTo(fc[0], fc[1]);
        //canvas.lineTo(tc[0], tc[1]);
        canvas.quadraticCurveTo(tc[0]-15, tc[1]-15, tc[0], tc[1]);
        canvas.stroke();
    }

    private void drawWord(double[] coords, double radius, final WordDTO dto, Color c) {
        int textLength = dto.getDisplayString().length();
        Label l = new Label(dto.getDisplayString());
        l.setStylePrimaryName(LABEL_STYLE);
        int x = new Double(coords[0]).intValue()-(6*(textLength/2));
        int y = new Double(coords[1]).intValue()-5;
        absolutePanel.add(l, x, y);
        coordinates.put(dto.getId(), coords);
        labels.add(l);
        registrations.add(l.addClickHandler(new ClickHandler() {
            
            @Override
            public void onClick(ClickEvent event) {
                placeService.changePlace(new WordDetailPlace(dto));
            }
        }));
        if (c != null) {
            // need to draw a box
            canvas.beginPath();
            canvas.setFillStyle(c);
            double width = l.getOffsetWidth();
            double height = l.getOffsetHeight();
            canvas.fillRect(x-2, y-2, width+4, height+4);
        }
    }

}
