package org.recursed.w.server.sql;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.recursed.w.shared.WordDTO;
import org.recursed.w.shared.headlines.HeadlineDTO;
import org.recursed.w.shared.search.SuggestedWordsResponse;

import com.google.inject.Inject;
import com.google.inject.Singleton;

@Singleton
public class WordsDAO {

    private static final String GET_WORDS_IN_RANGE = "select w.word_id, w.normal, sum(wh.frequency) as count "
            + "from words w, articles a, word_histogram wh where "
            + "w.ignore = 0 and "
            + "a.analyzed_on > ? and "
            + "a.analyzed_on < ? and "
            + "wh.article_id = a.article_id and "
            + "wh.word_id = w.word_id "
            + "group by wh.word_id order by count desc " + "limit ?";

    private static final String GET_WORD = "select * from words w where w.word_id = ?";

    private static final String GET_RELATED_WORDS_TIMELINE = "select w.word_id, w.normal, sum(wh2.frequency) wcount, "
            + "date(a.analyzed_on) from words w, word_histogram wh, "
            + "word_histogram wh2, articles a "
            + "where wh.word_id = ? and "
            + "wh.article_id = a.article_id "
            + "and wh2.article_id = wh.article_id "
            + "and w.word_id = wh2.word_id "
            + "and w.ignore = 0 and wh2.word_id != wh.word_id "
            + "and a.analyzed_on > ?  "
            + "group by w.word_id, 4 order by 4 asc ";

    private static final String GET_HEADLINES_FOR_WORD = "select s.source_id, a.title, a.url from "
            + "sources s, articles a, word_histogram wh where "
            + "s.source_id = a.source_id and "
            + "a.article_id = wh.article_id and "
            + "a.analyzed_on > ? and wh.word_id = ? " 
            + "order by a.analyzed_on";

    private static final String GET_SUGGESTED_WORDS = "select w.word_id, w.normal "
            + "from words w where w.normal like ? "
            + "and w.ignore = 0 "
            + "order by w.normal limit 20";
    
    private Pool pool;

    @Inject
    public WordsDAO(Pool pool) {
        this.pool = pool;
    }

    public List<WordDTO> getWords(int topN, long start, long finish) {
        Connection con = null;
        PreparedStatement st = null;
        ResultSet rs = null;
        Timestamp startDate = new Timestamp(start);
        Timestamp finishDate = new Timestamp(finish);
        ArrayList<WordDTO> result = new ArrayList<WordDTO>();
        long t1 = System.currentTimeMillis();
        try {
            con = pool.getConnection();
            st = con.prepareStatement(GET_WORDS_IN_RANGE);
            st.setTimestamp(1, startDate);
            st.setTimestamp(2, finishDate);
            st.setInt(3, topN);
            rs = st.executeQuery();
            while (rs.next()) {
                int wordId = rs.getInt(1);
                String label = rs.getString(2);
                int count = rs.getInt(3);
                WordDTO word = new WordDTO(wordId, label, count);
                result.add(word);
            }
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } finally {
            Pool.close(con, st, rs);
            long t2 = System.currentTimeMillis();
            System.out.println(generateLogLine("GET_WORDS_IN_RANGE", topN, start, finish, (t2-t1)));
        }
        return result;
    }

    public WordDTO getWord(int id) {
        Connection con = null;
        PreparedStatement st = null;
        ResultSet rs = null;
        WordDTO result;
        long t1 = System.currentTimeMillis();
        try {
            con = pool.getConnection();
            st = con.prepareStatement(GET_WORD);
            st.setInt(1, id);
            rs = st.executeQuery();
            if (rs.next()) {
                String normal = rs.getString("normal");
                result = new WordDTO(id, normal, 0);
            } else {
                throw new RuntimeException("Word not found with id: " + id);
            }
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } finally {
            Pool.close(con, st, rs);
            long t2 = System.currentTimeMillis();
            System.out.println(generateLogLine("GET_WORD", id, (t2-t1)));
        }
        return result;
    }

    public RelatedWordResult getRelatedWords(int id, long start) {
        Connection con = null;
        PreparedStatement st = null;
        ResultSet rs = null;
        Timestamp startDate = new Timestamp(start);
        RelatedWordResult result = new RelatedWordResult(new Date(start));
        long t1 = System.currentTimeMillis();
        try {
            con = pool.getConnection();
            st = con.prepareStatement(GET_RELATED_WORDS_TIMELINE);
            st.setInt(1, id);
            st.setTimestamp(2, startDate);
            rs = st.executeQuery();
            while (rs.next()) {
                int wordId = rs.getInt(1);
                String label = rs.getString(2);
                int count = rs.getInt(3);
                Date d = rs.getDate(4);
                WordDTO word = new WordDTO(wordId, label, count);
                result.addWord(word, d);
            }
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } finally {
            Pool.close(con, st, rs);
            long t2 = System.currentTimeMillis();
            System.out.println(generateLogLine("GET_RELATED_WORDS_TIMELINE", id, start, (t2-t1)));
        }
        return result;
    }

    public ArrayList<HeadlineDTO> getHeadlines(int wordId, long start) {
        ArrayList<HeadlineDTO> result = new ArrayList<HeadlineDTO>();
        Connection con = null;
        PreparedStatement st = null;
        ResultSet rs = null;
        Timestamp startDate = new Timestamp(start);
        long t1 = System.currentTimeMillis();
        try {
            con = pool.getConnection();
            st = con.prepareStatement(GET_HEADLINES_FOR_WORD);
            st.setTimestamp(1, startDate);
            st.setInt(2, wordId);
            rs = st.executeQuery();
            while (rs.next()) {
                int sourceId = rs.getInt(1);
                String title = rs.getString(2);
                String url = rs.getString(3);
                HeadlineDTO dto = new HeadlineDTO(title, url, sourceId);
                result.add(dto);
            }
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } finally {
            Pool.close(con, st, rs);
            long t2 = System.currentTimeMillis();
            System.out.println(generateLogLine("GET_HEADLINES_FOR_WORD", wordId, start, (t2-t1)));
        }
        return result;
    }

    public SuggestedWordsResponse getSuggestions(String segment) {
        Connection con = null;
        PreparedStatement st = null;
        ResultSet rs = null;
        SuggestedWordsResponse response = new SuggestedWordsResponse();
        long t1 = System.currentTimeMillis();
        try {
            con = pool.getConnection();
            st = con.prepareStatement(GET_SUGGESTED_WORDS);
            st.setString(1, segment + "%");
            rs = st.executeQuery();
            while (rs.next()) {
                int wordID = rs.getInt(1);
                String normal = rs.getString(2);
                WordDTO dto = new WordDTO(wordID, normal, -1);
                response.addSuggestion(dto);
            }
        } catch (SQLException e) {
            throw new RuntimeException(e);
        } finally {
            Pool.close(con, st, rs);
            long t2 = System.currentTimeMillis();
            System.out.println(generateLogLine("GET_SUGGESTED_WORDS", segment, (t2-t1)));
        }
        return response;
    }


    private static String generateLogLine(Object ...objects) {
        StringBuilder sb = new StringBuilder("SQL");
        for (Object o : objects) {
            sb.append("|");
            if (o != null) {
                sb.append(o);
            } else {
                sb.append("null");
            }
        }
        return sb.toString();
    }
}
